import time
from abc import ABC, abstractmethod


class RetryMechanism(ABC):
    @abstractmethod
    def is_done(self) -> bool:
        pass

    @abstractmethod
    def wait(self) -> None:
        pass

    @abstractmethod
    def step(self) -> None:
        pass


class MaxRetries(RetryMechanism):
    """Check for submissions status every 100 ms and retry a maximum of
    `max_retries` times."""

    def __init__(self, max_retries: int = 20):
        self.n_retries = 0
        self.max_retries = max_retries

    def step(self):
        self.n_retries += 1

    def wait(self):
        time.sleep(0.1)

    def is_done(self) -> bool:
        return self.n_retries >= self.max_retries


class MaxWaitTime(RetryMechanism):
    """Check for submissions status every 100 ms and wait for all submissions
    a maximum of `max_wait_time` (seconds)."""

    def __init__(self, max_wait_time_sec: float = 5 * 60):
        self.max_wait_time_sec = max_wait_time_sec
        self.total_wait_time = 0

    def step(self):
        self.total_wait_time += 0.1

    def wait(self):
        time.sleep(0.1)

    def is_done(self):
        return self.total_wait_time >= self.max_wait_time_sec


class RegularPeriodRetry(RetryMechanism):
    """Check for submissions status periodically for indefinite amount of time."""

    def __init__(self, wait_time_sec: float = 0.1):
        self.wait_time_sec = wait_time_sec

    def step(self):
        pass

    def wait(self):
        time.sleep(self.wait_time_sec)

    def is_done(self) -> bool:
        return False



from typing import Iterable, Union

import requests

from .base_types import Config, Language, LanguageAlias
from .data import LANGUAGE_TO_LANGUAGE_ID
from .submission import Submission, Submissions


class Client:
    API_KEY_ENV = "JUDGE0_API_KEY"
    DEFAULT_MAX_SUBMISSION_BATCH_SIZE = 20
    ENABLED_BATCHED_SUBMISSIONS = True
    EFFECTIVE_SUBMISSION_BATCH_SIZE = (
        DEFAULT_MAX_SUBMISSION_BATCH_SIZE if ENABLED_BATCHED_SUBMISSIONS else 1
    )

    def __init__(self, endpoint, auth_headers) -> None:
        self.endpoint = endpoint
        self.auth_headers = auth_headers

        try:
            self.languages = [Language(**lang) for lang in self.get_languages()]
            self.config = Config(**self.get_config_info())
        except Exception as e:
            raise RuntimeError(
                f"Authentication failed. Visit {self.HOME_URL} to get or review your authentication credentials."
            ) from e

    def get_about(self) -> dict:
        r = requests.get(
            f"{self.endpoint}/about",
            headers=self.auth_headers,
        )
        r.raise_for_status()
        return r.json()

    def get_config_info(self) -> dict:
        r = requests.get(
            f"{self.endpoint}/config_info",
            headers=self.auth_headers,
        )
        r.raise_for_status()
        return r.json()

    def get_language(self, language_id) -> dict:
        request_url = f"{self.endpoint}/languages/{language_id}"
        r = requests.get(request_url, headers=self.auth_headers)
        r.raise_for_status()
        return r.json()

    def get_languages(self) -> list[dict]:
        request_url = f"{self.endpoint}/languages"
        r = requests.get(request_url, headers=self.auth_headers)
        r.raise_for_status()
        return r.json()

    def get_statuses(self) -> list[dict]:
        r = requests.get(
            f"{self.endpoint}/statuses",
            headers=self.auth_headers,
        )
        r.raise_for_status()
        return r.json()

    @property
    def version(self):
        if not hasattr(self, "_version"):
            _version = self.get_about()["version"]
            setattr(self, "_version", _version)
        return self._version

    def get_language_id(self, language: Union[LanguageAlias, int]) -> int:
        """Get language id for the corresponding language alias for the client."""
        if isinstance(language, LanguageAlias):
            supported_language_ids = LANGUAGE_TO_LANGUAGE_ID[self.version]
            language = supported_language_ids.get(language, -1)
        return language

    def is_language_supported(self, language: Union[LanguageAlias, int]) -> bool:
        """Check if language is supported by the client."""
        language_id = self.get_language_id(language)
        return any(language_id == lang.id for lang in self.languages)

    def create_submission(self, submission: Submission) -> Submission:
        # Check if the client supports the language specified in the submission.
        if not self.is_language_supported(language=submission.language):
            raise RuntimeError(
                f"Client {type(self).__name__} does not support language with "
                f"id {submission.language}!"
            )

        params = {
            "base64_encoded": "true",
            "wait": "false",
        }

        body = submission.as_body(self)

        resp = requests.post(
            f"{self.endpoint}/submissions",
            json=body,
            params=params,
            headers=self.auth_headers,
        )
        resp.raise_for_status()

        submission.set_attributes(resp.json())

        return submission

    def get_submission(
        self,
        submission: Submission,
        *,
        fields: Union[str, Iterable[str], None] = None,
    ) -> Submission:
        """Check the submission status."""

        params = {
            "base64_encoded": "true",
        }

        if isinstance(fields, str):
            fields = [fields]

        if fields is not None:
            params["fields"] = ",".join(fields)
        else:
            params["fields"] = "*"

        resp = requests.get(
            f"{self.endpoint}/submissions/{submission.token}",
            params=params,
            headers=self.auth_headers,
        )
        resp.raise_for_status()

        submission.set_attributes(resp.json())

        return submission

    def create_submissions(self, submissions: Submissions) -> Submissions:
        # Check if all submissions contain supported language.
        for submission in submissions:
            if not self.is_language_supported(language=submission.language):
                raise RuntimeError(
                    f"Client {type(self).__name__} does not support language "
                    f"{submission.language}!"
                )

        submissions_body = [submission.as_body(self) for submission in submissions]

        resp = requests.post(
            f"{self.endpoint}/submissions/batch",
            headers=self.auth_headers,
            params={"base64_encoded": "true"},
            json={"submissions": submissions_body},
        )
        resp.raise_for_status()

        for submission, attrs in zip(submissions, resp.json()):
            submission.set_attributes(attrs)

        return submissions

    def get_submissions(
        self,
        submissions: Submissions,
        *,
        fields: Union[str, Iterable[str], None] = None,
    ) -> Submissions:
        params = {
            "base64_encoded": "true",
        }

        if isinstance(fields, str):
            fields = [fields]

        if fields is not None:
            params["fields"] = ",".join(fields)
        else:
            params["fields"] = "*"

        tokens = ",".join(submission.token for submission in submissions)
        params["tokens"] = tokens

        resp = requests.get(
            f"{self.endpoint}/submissions/batch",
            params=params,
            headers=self.auth_headers,
        )
        resp.raise_for_status()

        for submission, attrs in zip(submissions, resp.json()["submissions"]):
            submission.set_attributes(attrs)

        return submissions


class ATD(Client):
    API_KEY_ENV = "JUDGE0_ATD_API_KEY"

    def __init__(self, endpoint, host_header_value, api_key):
        self.api_key = api_key
        super().__init__(
            endpoint,
            {
                "x-apihub-host": host_header_value,
                "x-apihub-key": api_key,
            },
        )

    def _update_endpoint_header(self, header_value):
        self.auth_headers["x-apihub-endpoint"] = header_value


class ATDJudge0CE(ATD):
    DEFAULT_ENDPOINT: str = "https://judge0-ce.proxy-production.allthingsdev.co"
    DEFAULT_HOST: str = "Judge0-CE.allthingsdev.co"
    HOME_URL: str = (
        "https://www.allthingsdev.co/apimarketplace/judge0-ce/66b683c8b7b7ad054eb6ff8f"
    )

    DEFAULT_ABOUT_ENDPOINT: str = "01fc1c98-ceee-4f49-8614-f2214703e25f"
    DEFAULT_CONFIG_INFO_ENDPOINT: str = "b7aab45d-5eb0-4519-b092-89e5af4fc4f3"
    DEFAULT_LANGUAGE_ENDPOINT: str = "a50ae6b1-23c1-40eb-b34c-88bc8cf2c764"
    DEFAULT_LANGUAGES_ENDPOINT: str = "03824deb-bd18-4456-8849-69d78e1383cc"
    DEFAULT_STATUSES_ENDPOINT: str = "c37b603f-6f99-4e31-a361-7154c734f19b"
    DEFAULT_CREATE_SUBMISSION_ENDPOINT: str = "6e65686d-40b0-4bf7-a12f-1f6d033c4473"
    DEFAULT_GET_SUBMISSION_ENDPOINT: str = "b7032b8b-86da-40b4-b9d3-b1f5e2b4ee1e"
    DEFAULT_CREATE_SUBMISSIONS_ENDPOINT: str = "402b857c-1126-4450-bfd8-22e1f2cbff2f"
    DEFAULT_GET_SUBMISSIONS_ENDPOINT: str = "e42f2a26-5b02-472a-80c9-61c4bdae32ec"

    def __init__(self, api_key):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
        )

    def get_about(self) -> dict:
        self._update_endpoint_header(self.DEFAULT_ABOUT_ENDPOINT)
        return super().get_about()

    def get_config_info(self) -> dict:
        self._update_endpoint_header(self.DEFAULT_CONFIG_INFO_ENDPOINT)
        return super().get_config_info()

    def get_language(self, language_id) -> dict:
        self._update_endpoint_header(self.DEFAULT_LANGUAGE_ENDPOINT)
        return super().get_language(language_id)

    def get_languages(self) -> list[dict]:
        self._update_endpoint_header(self.DEFAULT_LANGUAGES_ENDPOINT)
        return super().get_languages()

    def get_statuses(self) -> list[dict]:
        self._update_endpoint_header(self.DEFAULT_STATUSES_ENDPOINT)
        return super().get_statuses()

    def create_submission(self, submission: Submission) -> Submission:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSION_ENDPOINT)
        return super().create_submission(submission)

    def get_submission(
        self,
        submission: Submission,
        *,
        fields: Union[str, Iterable[str], None] = None,
    ) -> Submission:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSION_ENDPOINT)
        return super().get_submission(submission, fields=fields)

    def create_submissions(self, submissions: Submissions) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSIONS_ENDPOINT)
        return super().create_submissions(submissions)

    def get_submissions(
        self,
        submissions: Submissions,
        *,
        fields: Union[str, Iterable[str], None] = None,
    ) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSIONS_ENDPOINT)
        return super().get_submissions(submissions, fields=fields)


class ATDJudge0ExtraCE(ATD):
    DEFAULT_ENDPOINT: str = "https://judge0-extra-ce.proxy-production.allthingsdev.co"
    DEFAULT_HOST: str = "Judge0-Extra-CE.allthingsdev.co"
    HOME_URL: str = (
        "https://www.allthingsdev.co/apimarketplace/judge0-extra-ce/66b68838b7b7ad054eb70690"
    )

    DEFAULT_ABOUT_ENDPOINT: str = "1fd631a1-be6a-47d6-bf4c-987e357e3096"
    DEFAULT_CONFIG_INFO_ENDPOINT: str = "46e05354-2a43-436a-9458-5d111456f0ff"
    DEFAULT_LANGUAGE_ENDPOINT: str = "10465a84-2a2c-4213-845f-45e3c04a5867"
    DEFAULT_LANGUAGES_ENDPOINT: str = "774ecece-1200-41f7-a992-38f186c90803"
    DEFAULT_STATUSES_ENDPOINT: str = "a2843b3c-673d-4966-9a14-2e7d76dcd0cb"
    DEFAULT_CREATE_SUBMISSION_ENDPOINT: str = "be2d195e-dd58-4770-9f3c-d6c0fbc2b6e5"
    DEFAULT_GET_SUBMISSION_ENDPOINT: str = "c3a457cd-37a6-4106-97a8-9e60a223abbc"
    DEFAULT_CREATE_SUBMISSIONS_ENDPOINT: str = "c64df5d3-edfd-4b08-8687-561af2f80d2f"
    DEFAULT_GET_SUBMISSIONS_ENDPOINT: str = "5d173718-8e6a-4cf5-9d8c-db5e6386d037"

    def __init__(self, api_key):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
        )

    def get_about(self) -> dict:
        self._update_endpoint_header(self.DEFAULT_ABOUT_ENDPOINT)
        return super().get_about()

    def get_config_info(self) -> dict:
        self._update_endpoint_header(self.DEFAULT_CONFIG_INFO_ENDPOINT)
        return super().get_config_info()

    def get_language(self, language_id) -> dict:
        self._update_endpoint_header(self.DEFAULT_LANGUAGE_ENDPOINT)
        return super().get_language(language_id)

    def get_languages(self) -> list[dict]:
        self._update_endpoint_header(self.DEFAULT_LANGUAGES_ENDPOINT)
        return super().get_languages()

    def get_statuses(self) -> list[dict]:
        self._update_endpoint_header(self.DEFAULT_STATUSES_ENDPOINT)
        return super().get_statuses()

    def create_submission(self, submission: Submission) -> Submission:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSION_ENDPOINT)
        return super().create_submission(submission)

    def get_submission(
        self,
        submission: Submission,
        *,
        fields: Union[str, Iterable[str], None] = None,
    ) -> Submission:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSION_ENDPOINT)
        return super().get_submission(submission, fields=fields)

    def create_submissions(self, submissions: Submissions) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_CREATE_SUBMISSIONS_ENDPOINT)
        return super().create_submissions(submissions)

    def get_submissions(
        self,
        submissions: Submissions,
        *,
        fields: Union[str, Iterable[str], None] = None,
    ) -> Submissions:
        self._update_endpoint_header(self.DEFAULT_GET_SUBMISSIONS_ENDPOINT)
        return super().get_submissions(submissions, fields=fields)


class Rapid(Client):
    API_KEY_ENV = "JUDGE0_RAPID_API_KEY"

    def __init__(self, endpoint, host_header_value, api_key):
        self.api_key = api_key
        super().__init__(
            endpoint,
            {
                "x-rapidapi-host": host_header_value,
                "x-rapidapi-key": api_key,
            },
        )


class RapidJudge0CE(Rapid):
    DEFAULT_ENDPOINT: str = "https://judge0-ce.p.rapidapi.com"
    DEFAULT_HOST: str = "judge0-ce.p.rapidapi.com"
    HOME_URL: str = "https://rapidapi.com/judge0-official/api/judge0-ce"

    def __init__(self, api_key):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
        )


class RapidJudge0ExtraCE(Rapid):
    DEFAULT_ENDPOINT: str = "https://judge0-extra-ce.p.rapidapi.com"
    DEFAULT_HOST: str = "judge0-extra-ce.p.rapidapi.com"
    HOME_URL: str = "https://rapidapi.com/judge0-official/api/judge0-extra-ce"

    def __init__(self, api_key):
        super().__init__(
            self.DEFAULT_ENDPOINT,
            self.DEFAULT_HOST,
            api_key,
        )


class Sulu(Client):
    API_KEY_ENV = "JUDGE0_SULU_API_KEY"

    def __init__(self, endpoint, api_key=None):
        self.api_key = api_key
        super().__init__(
            endpoint,
            {"Authorization": f"Bearer {api_key}"} if api_key else None,
        )


class SuluJudge0CE(Sulu):
    DEFAULT_ENDPOINT: str = "https://judge0-ce.p.sulu.sh"
    HOME_URL: str = "https://sparkhub.sulu.sh/apis/judge0/judge0-ce/readme"

    def __init__(self, api_key=None):
        super().__init__(self.DEFAULT_ENDPOINT, api_key)


class SuluJudge0ExtraCE(Sulu):
    DEFAULT_ENDPOINT: str = "https://judge0-extra-ce.p.sulu.sh"
    HOME_URL: str = "https://sparkhub.sulu.sh/apis/judge0/judge0-extra-ce/readme"

    def __init__(self, api_key=None):
        super().__init__(self.DEFAULT_ENDPOINT, api_key)


CE = [RapidJudge0CE, SuluJudge0CE, ATDJudge0CE]
EXTRA_CE = [RapidJudge0ExtraCE, SuluJudge0ExtraCE, ATDJudge0ExtraCE]



import os

from .api import (
    async_execute,
    async_run,
    execute,
    get_client,
    run,
    sync_execute,
    sync_run,
    wait,
)
from .base_types import Flavor, Language, LanguageAlias, Status, TestCase
from .clients import (
    ATD,
    ATDJudge0CE,
    ATDJudge0ExtraCE,
    Client,
    Rapid,
    RapidJudge0CE,
    RapidJudge0ExtraCE,
    Sulu,
    SuluJudge0CE,
    SuluJudge0ExtraCE,
)
from .filesystem import File, Filesystem
from .retry import MaxRetries, MaxWaitTime, RegularPeriodRetry
from .submission import Submission

__all__ = [
    "ATD",
    "ATDJudge0CE",
    "ATDJudge0ExtraCE",
    "Client",
    "File",
    "Filesystem",
    "Language",
    "LanguageAlias",
    "MaxRetries",
    "MaxWaitTime",
    "Rapid",
    "RapidJudge0CE",
    "RapidJudge0ExtraCE",
    "RegularPeriodRetry",
    "Status",
    "Submission",
    "Sulu",
    "SuluJudge0CE",
    "SuluJudge0ExtraCE",
    "TestCase",
    "async_execute",
    "execute",
    "get_client",
    "async_run",
    "sync_run",
    "run",
    "sync_execute",
    "wait",
]

JUDGE0_IMPLICIT_CE_CLIENT = None
JUDGE0_IMPLICIT_EXTRA_CE_CLIENT = None


def _get_implicit_client(flavor: Flavor) -> Client:
    global JUDGE0_IMPLICIT_CE_CLIENT, JUDGE0_IMPLICIT_EXTRA_CE_CLIENT

    # Implicit clients are already set.
    if flavor == Flavor.CE and JUDGE0_IMPLICIT_CE_CLIENT is not None:
        return JUDGE0_IMPLICIT_CE_CLIENT
    if flavor == Flavor.EXTRA_CE and JUDGE0_IMPLICIT_EXTRA_CE_CLIENT is not None:
        return JUDGE0_IMPLICIT_EXTRA_CE_CLIENT

    from .clients import CE, EXTRA_CE

    try:
        from dotenv import load_dotenv

        load_dotenv()
    except:  # noqa: E722
        pass

    if flavor == Flavor.CE:
        client_classes = CE
    else:
        client_classes = EXTRA_CE

    # Try to find one of the predefined keys JUDGE0_{SULU,RAPID,ATD}_API_KEY
    # in environment variables.
    client = None
    for client_class in client_classes:
        api_key = os.getenv(client_class.API_KEY_ENV)
        if api_key is not None:
            client = client_class(api_key)
            break

    # If we didn't find any of the possible predefined keys, initialize
    # the preview Sulu client based on the flavor.
    if client is None:
        if flavor == Flavor.CE:
            client = SuluJudge0CE()
        else:
            client = SuluJudge0ExtraCE()

    if flavor == Flavor.CE:
        JUDGE0_IMPLICIT_CE_CLIENT = client
    else:
        JUDGE0_IMPLICIT_EXTRA_CE_CLIENT = client

    return client


CE = Flavor.CE
EXTRA_CE = Flavor.EXTRA_CE

PYTHON = LanguageAlias.PYTHON
CPP = LanguageAlias.CPP
JAVA = LanguageAlias.JAVA
CPP_GCC = LanguageAlias.CPP_GCC
CPP_CLANG = LanguageAlias.CPP_CLANG
PYTHON_FOR_ML = LanguageAlias.PYTHON_FOR_ML



from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, Union


TestCases = Union[
    list["TestCase"],
    tuple["TestCase"],
    list[dict],
    tuple[dict],
    list[list],
    list[tuple],
    tuple[list],
    tuple[tuple],
]


@dataclass(frozen=True)
class TestCase:
    # Needed to disable pytest from recognizing it as a class containing different test cases.
    __test__ = False

    input: Optional[str] = None
    expected_output: Optional[str] = None

    @staticmethod
    def from_record(
        test_case: Optional[Union[tuple, list, dict, "TestCase"]] = None
    ) -> "TestCase":
        if isinstance(test_case, (tuple, list)):
            test_case = {
                field: value
                for field, value in zip(("input", "expected_output"), test_case)
            }
        if isinstance(test_case, dict):
            return TestCase(
                input=test_case.get("input", None),
                expected_output=test_case.get("expected_output", None),
            )
        if isinstance(test_case, TestCase) or test_case is None:
            return test_case
        raise ValueError(
            f"Cannot create TestCase object from object of type {type(test_case)}."
        )


class Encodeable(ABC):
    @abstractmethod
    def encode(self) -> bytes:
        pass


@dataclass(frozen=True)
class Language:
    id: int
    name: str


class LanguageAlias(IntEnum):
    PYTHON = 0
    CPP = 1
    JAVA = 2
    CPP_GCC = 3
    CPP_CLANG = 4
    PYTHON_FOR_ML = 5


class Flavor(IntEnum):
    CE = 0
    EXTRA_CE = 1


class Status(IntEnum):
    IN_QUEUE = 1
    PROCESSING = 2
    ACCEPTED = 3
    WRONG_ANSWER = 4
    TIME_LIMIT_EXCEEDED = 5
    COMPILATION_ERROR = 6
    RUNTIME_ERROR_SIGSEGV = 7
    RUNTIME_ERROR_SIGXFSZ = 8
    RUNTIME_ERROR_SIGFPE = 9
    RUNTIME_ERROR_SIGABRT = 10
    RUNTIME_ERROR_NZEC = 11
    RUNTIME_ERROR_OTHER = 12
    INTERNAL_ERROR = 13
    EXEC_FORMAT_ERROR = 14

    def __str__(self):
        return self.name.lower().replace("_", " ").title()


@dataclass(frozen=True)
class Config:
    allow_enable_network: bool
    allow_enable_per_process_and_thread_memory_limit: bool
    allow_enable_per_process_and_thread_time_limit: bool
    allowed_languages_for_compile_options: list[str]
    callbacks_max_tries: int
    callbacks_timeout: float
    cpu_extra_time: float
    cpu_time_limit: float
    enable_additional_files: bool
    enable_batched_submissions: bool
    enable_callbacks: bool
    enable_command_line_arguments: bool
    enable_compiler_options: bool
    enable_network: bool
    enable_per_process_and_thread_memory_limit: bool
    enable_per_process_and_thread_time_limit: bool
    enable_submission_delete: bool
    enable_wait_result: bool
    maintenance_mode: bool
    max_cpu_extra_time: float
    max_cpu_time_limit: float
    max_extract_size: int
    max_file_size: int
    max_max_file_size: int
    max_max_processes_and_or_threads: int
    max_memory_limit: int
    max_number_of_runs: int
    max_processes_and_or_threads: int
    max_queue_size: int
    max_stack_limit: int
    max_submission_batch_size: int
    max_wall_time_limit: float
    memory_limit: int
    number_of_runs: int
    redirect_stderr_to_stdout: bool
    stack_limit: int
    submission_cache_duration: float
    use_docs_as_homepage: bool
    wall_time_limit: float



from .base_types import LanguageAlias

LANGUAGE_TO_LANGUAGE_ID = {
    "1.13.1": {
        LanguageAlias.PYTHON: 71,
        LanguageAlias.CPP: 54,
        LanguageAlias.JAVA: 62,
        LanguageAlias.CPP_GCC: 54,
        LanguageAlias.CPP_CLANG: 76,
    },
    "1.13.1-extra": {
        LanguageAlias.PYTHON: 10,
        LanguageAlias.CPP: 2,
        LanguageAlias.JAVA: 4,
        LanguageAlias.CPP_CLANG: 2,
        LanguageAlias.PYTHON_FOR_ML: 10,
    },
    "1.14.0": {
        LanguageAlias.PYTHON: 100,
        LanguageAlias.CPP: 105,
        LanguageAlias.JAVA: 91,
        LanguageAlias.CPP_GCC: 105,
        LanguageAlias.CPP_CLANG: 76,
    },
    "1.14.0-extra": {
        LanguageAlias.PYTHON: 25,
        LanguageAlias.CPP: 2,
        LanguageAlias.JAVA: 4,
        LanguageAlias.CPP_CLANG: 2,
        LanguageAlias.PYTHON_FOR_ML: 25,
    },
}



import copy
from datetime import datetime
from typing import Optional, Union

from judge0.filesystem import Filesystem

from .base_types import LanguageAlias, Status
from .common import decode, encode

ENCODED_REQUEST_FIELDS = {
    "source_code",
    "additional_files",
    "stdin",
    "expected_output",
}
ENCODED_RESPONSE_FIELDS = {
    "stdout",
    "stderr",
    "compile_output",
    # "post_execution_filesystem",
}
ENCODED_FIELDS = ENCODED_REQUEST_FIELDS | ENCODED_RESPONSE_FIELDS
EXTRA_REQUEST_FIELDS = {
    "compiler_options",
    "command_line_arguments",
    "cpu_time_limit",
    "cpu_extra_time",
    "wall_time_limit",
    "memory_limit",
    "stack_limit",
    "max_processes_and_or_threads",
    "enable_per_process_and_thread_time_limit",
    "enable_per_process_and_thread_memory_limit",
    "max_file_size",
    "redirect_stderr_to_stdout",
    "enable_network",
    "number_of_runs",
    "callback_url",
}
EXTRA_RESPONSE_FIELDS = {
    "message",
    "exit_code",
    "exit_signal",
    "status",
    "created_at",
    "finished_at",
    "token",
    "time",
    "wall_time",
    "memory",
    "post_execution_filesystem",
}
REQUEST_FIELDS = ENCODED_REQUEST_FIELDS | EXTRA_REQUEST_FIELDS
RESPONSE_FIELDS = ENCODED_RESPONSE_FIELDS | EXTRA_RESPONSE_FIELDS
FIELDS = REQUEST_FIELDS | RESPONSE_FIELDS
SKIP_FIELDS = {"language_id", "language", "status_id"}
DATETIME_FIELDS = {"created_at", "finished_at"}
FLOATING_POINT_FIELDS = {
    "cpu_time_limit",
    "cpu_extra_time",
    "time",
    "wall_time",
    "wall_time_limit",
}

Submissions = Union[list["Submission"], tuple["Submission"]]


class Submission:
    """
    Stores a representation of a Submission to/from Judge0.
    """

    def __init__(
        self,
        *,
        source_code: Optional[str] = None,
        language: Union[LanguageAlias, int] = LanguageAlias.PYTHON,
        additional_files=None,
        compiler_options=None,
        command_line_arguments=None,
        stdin=None,
        expected_output=None,
        cpu_time_limit=None,
        cpu_extra_time=None,
        wall_time_limit=None,
        memory_limit=None,
        stack_limit=None,
        max_processes_and_or_threads=None,
        enable_per_process_and_thread_time_limit=None,
        enable_per_process_and_thread_memory_limit=None,
        max_file_size=None,
        redirect_stderr_to_stdout=None,
        enable_network=None,
        number_of_runs=None,
        callback_url=None,
    ):
        self.source_code = source_code
        self.language = language
        self.additional_files = additional_files

        # Extra pre-execution submission attributes.
        self.compiler_options = compiler_options
        self.command_line_arguments = command_line_arguments
        self.stdin = stdin
        self.expected_output = expected_output
        self.cpu_time_limit = cpu_time_limit
        self.cpu_extra_time = cpu_extra_time
        self.wall_time_limit = wall_time_limit
        self.memory_limit = memory_limit
        self.stack_limit = stack_limit
        self.max_processes_and_or_threads = max_processes_and_or_threads
        self.enable_per_process_and_thread_time_limit = (
            enable_per_process_and_thread_time_limit
        )
        self.enable_per_process_and_thread_memory_limit = (
            enable_per_process_and_thread_memory_limit
        )
        self.max_file_size = max_file_size
        self.redirect_stderr_to_stdout = redirect_stderr_to_stdout
        self.enable_network = enable_network
        self.number_of_runs = number_of_runs
        self.callback_url = callback_url

        # Post-execution submission attributes.
        self.stdout = None
        self.stderr = None
        self.compile_output = None
        self.message = None
        self.exit_code = None
        self.exit_signal = None
        self.status = None
        self.created_at = None
        self.finished_at = None
        self.token = ""
        self.time = None
        self.wall_time = None
        self.memory = None
        self.post_execution_filesystem = None

    def set_attributes(self, attributes):
        for attr, value in attributes.items():
            if attr in SKIP_FIELDS:
                continue

            if attr in ENCODED_FIELDS:
                value = decode(value) if value else None
            elif attr == "status":
                value = Status(value["id"])
            elif attr in DATETIME_FIELDS and value is not None:
                value = datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")
            elif attr in FLOATING_POINT_FIELDS and value is not None:
                value = float(value)
            elif attr == "post_execution_filesystem":
                value = Filesystem(value)

            setattr(self, attr, value)

    def as_body(self, client: "Client") -> dict:
        body = {
            "source_code": encode(self.source_code),
            "language_id": client.get_language_id(self.language),
        }

        for field in ENCODED_REQUEST_FIELDS:
            value = getattr(self, field)
            if value is not None:
                body[field] = encode(value)

        for field in EXTRA_REQUEST_FIELDS:
            value = getattr(self, field)
            if value is not None:
                body[field] = value

        return body

    def is_done(self) -> bool:
        if self.status is None:
            return False
        else:
            return self.status not in (Status.IN_QUEUE, Status.PROCESSING)

    def pre_execution_copy(self) -> "Submission":
        new_submission = Submission()
        for attr in REQUEST_FIELDS:
            setattr(new_submission, attr, copy.deepcopy(getattr(self, attr)))
        return new_submission

    def __repr__(self) -> str:
        arguments = ", ".join(f"{field}={getattr(self, field)!r}" for field in FIELDS)
        return f"{self.__class__.__name__}({arguments})"

    def __iter__(self):
        if self.post_execution_filesystem is None:
            return iter([])
        else:
            return iter(self.post_execution_filesystem)



from judge0 import Status, Submission, wait


def test_status_before_and_after_submission(request):
    client = request.getfixturevalue("judge0_ce_client")
    submission = Submission(source_code='print("Hello World!")')

    assert submission.status is None

    client.create_submission(submission)
    client.get_submission(submission)

    assert submission.status.__class__ == Status
    assert submission.status >= Status.IN_QUEUE


def test_is_done(request):
    client = request.getfixturevalue("judge0_ce_client")
    submission = Submission(source_code='print("Hello World!")')

    assert submission.status is None

    client.create_submission(submission)
    wait(client=client, submissions=submission)

    assert submission.is_done()



import copy
import io
import zipfile

from base64 import b64decode, b64encode
from collections import abc
from typing import Iterable, Optional, Union

from .base_types import Encodeable


class File:
    def __init__(self, name: str, content: Optional[Union[str, bytes]] = None):
        self.name = name

        # Let's keep content attribute internally encoded as bytes.
        if isinstance(content, str):
            self.content = content.encode()
        elif isinstance(content, bytes):
            self.content = content
        else:
            self.content = b""

    def __str__(self):
        return self.content.decode(errors="backslashreplace")


class Filesystem(Encodeable):
    def __init__(
        self,
        content: Optional[Union[str, bytes, File, Iterable[File], "Filesystem"]] = None,
    ):
        self.files: list[File] = []

        if isinstance(content, (bytes, str)):
            if isinstance(content, bytes):
                zip_bytes = content
            else:
                zip_bytes = b64decode(content.encode())

            with zipfile.ZipFile(io.BytesIO(zip_bytes), "r") as zip_file:
                for file_name in zip_file.namelist():
                    with zip_file.open(file_name) as fp:
                        self.files.append(File(file_name, fp.read()))
        elif isinstance(content, abc.Iterable):
            self.files = list(content)
        elif isinstance(content, File):
            self.files = [content]
        elif isinstance(content, Filesystem):
            self.files = copy.deepcopy(content.files)

    def __repr__(self) -> str:
        content_encoded = b64encode(self.encode()).decode()
        return f"{self.__class__.__name__}(content={content_encoded!r})"

    def encode(self) -> bytes:
        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w") as zip_file:
            for file in self.files:
                zip_file.writestr(file.name, file.content)
        return zip_buffer.getvalue()

    def __str__(self) -> str:
        return b64encode(self.encode()).decode()

    def __iter__(self):
        return iter(self.files)



from base64 import b64decode, b64encode
from itertools import islice
from typing import Union

from .base_types import Encodeable


def encode(content: Union[bytes, str, Encodeable]) -> str:
    if isinstance(content, bytes):
        return b64encode(content).decode()
    if isinstance(content, str):
        return b64encode(content.encode()).decode()
    if isinstance(content, Encodeable):
        return b64encode(content.encode()).decode()
    raise ValueError(f"Unsupported type. Expected bytes or str, got {type(content)}!")


def decode(content: Union[bytes, str]) -> str:
    if isinstance(content, bytes):
        return b64decode(content.decode(errors="backslashreplace")).decode(
            errors="backslashreplace"
        )
    if isinstance(content, str):
        return b64decode(content.encode()).decode(errors="backslashreplace")
    raise ValueError(f"Unsupported type. Expected bytes or str, got {type(content)}!")


def batched(iterable, n):
    """Utility function for batching submissions.

    Adapted from https://docs.python.org/3/library/itertools.html#itertools.batched.
    """
    if n < 1:
        raise ValueError("n must be at least one")
    iterator = iter(iterable)
    while True:
        batch = tuple(islice(iterator, n))
        if not batch:
            break
        yield batch



from typing import Iterable, Optional, Union

from .base_types import Flavor, TestCase, TestCases
from .clients import Client
from .common import batched

from .retry import RegularPeriodRetry, RetryMechanism
from .submission import Submission, Submissions


def get_client(flavor: Flavor = Flavor.CE) -> Client:
    from . import _get_implicit_client

    if isinstance(flavor, Flavor):
        return _get_implicit_client(flavor=flavor)
    else:
        raise ValueError(
            "Expected argument flavor to be of of type enum Flavor, "
            f"got {type(flavor)}."
        )


def _resolve_client(
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
) -> Client:
    # User explicitly passed a client.
    if isinstance(client, Client):
        return client

    if isinstance(client, Flavor):
        return get_client(client)

    if client is None and isinstance(submissions, list) and len(submissions) == 0:
        raise ValueError("Client cannot be determined from empty submissions.")

    # client is None and we have to determine a flavor of the client from the
    # submissions and the languages.
    if isinstance(submissions, Submission):
        submissions = [submissions]

    # Check which client supports all languages from the provided submissions.
    languages = [submission.language for submission in submissions]

    for flavor in Flavor:
        client = get_client(flavor)
        if client is not None and all(
            (client.is_language_supported(lang) for lang in languages)
        ):
            return client

    raise RuntimeError(
        "Failed to resolve the client from submissions argument. "
        "None of the implicit clients supports all languages from the submissions. "
        "Please explicitly provide the client argument."
    )


def create_submissions(
    client: Optional[Client] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
) -> Union[Submission, Submissions]:
    client = _resolve_client(client=client, submissions=submissions)

    if isinstance(submissions, Submission):
        return client.create_submission(submissions)

    result_submissions = []
    for submission_batch in batched(
        submissions, client.config.max_submission_batch_size
    ):
        if len(submission_batch) > 1:
            result_submissions.extend(client.create_submissions(submission_batch))
        else:
            result_submissions.append(client.create_submission(submission_batch[0]))

    return result_submissions


def get_submissions(
    *,
    client: Optional[Client] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    fields: Union[str, Iterable[str], None] = None,
) -> Union[Submission, Submissions]:
    client = _resolve_client(client=client, submissions=submissions)

    if isinstance(submissions, Submission):
        return client.get_submission(submissions, fields=fields)

    result_submissions = []
    for submission_batch in batched(
        submissions, client.config.max_submission_batch_size
    ):
        if len(submission_batch) > 1:
            result_submissions.extend(
                client.get_submissions(submission_batch, fields=fields)
            )
        else:
            result_submissions.append(
                client.get_submission(submission_batch[0], fields=fields)
            )

    return result_submissions


def wait(
    *,
    client: Optional[Client] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    retry_mechanism: Optional[RetryMechanism] = None,
) -> Union[Submission, Submissions]:
    client = _resolve_client(client, submissions)

    if retry_mechanism is None:
        retry_mechanism = RegularPeriodRetry()

    if isinstance(submissions, Submission):
        submissions_to_check = {
            submission.token: submission for submission in [submissions]
        }
    else:
        submissions_to_check = {
            submission.token: submission for submission in submissions
        }

    while len(submissions_to_check) > 0 and not retry_mechanism.is_done():
        get_submissions(client=client, submissions=list(submissions_to_check.values()))
        for token in list(submissions_to_check):
            submission = submissions_to_check[token]
            if submission.is_done():
                submissions_to_check.pop(token)

        # Don't wait if there is no submissions to check for anymore.
        if len(submissions_to_check) == 0:
            break

        retry_mechanism.wait()
        retry_mechanism.step()

    return submissions


def create_submissions_from_test_cases(
    submissions: Union[Submission, Submissions],
    test_cases: Optional[Union[TestCase, TestCases]] = None,
):
    """Utility function for creating submissions from the (submission, test_case) pairs.

    The following table contains the return type based on the types of `submissions`
    and `test_cases` arguments:

    | submissions | test_cases | returns     |
    |:------------|:-----------|:------------|
    | Submission  | TestCase   | Submission  |
    | Submission  | TestCases  | Submissions |
    | Submissions | TestCase   | Submissions |
    | Submissions | TestCases  | Submissions |

    """
    if isinstance(submissions, Submission):
        submissions_list = [submissions]
    else:
        submissions_list = submissions

    if isinstance(test_cases, TestCase) or test_cases is None:
        test_cases_list = [test_cases]
    else:
        test_cases_list = test_cases

    test_cases_list = [TestCase.from_record(tc) for tc in test_cases_list]

    all_submissions = []
    for submission in submissions_list:
        for test_case in test_cases_list:
            submission_copy = submission.pre_execution_copy()
            if test_case is not None:
                submission_copy.stdin = test_case.input
                submission_copy.expected_output = test_case.expected_output
            all_submissions.append(submission_copy)

    if isinstance(submissions, Submission) and (
        isinstance(test_cases, TestCase) or test_cases is None
    ):
        return all_submissions[0]
    else:
        return all_submissions


def _execute(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    source_code: Optional[str] = None,
    test_cases: Optional[Union[TestCase, TestCases]] = None,
    wait_for_result: bool = False,
    **kwargs,
) -> Union[Submission, Submissions]:
    if submissions is not None and source_code is not None:
        raise ValueError(
            "Both submissions and source_code arguments are provided. "
            "Provide only one of the two."
        )
    if submissions is None and source_code is None:
        raise ValueError("Neither source_code nor submissions argument are provided.")

    if source_code is not None:
        submissions = Submission(source_code=source_code, **kwargs)

    client = _resolve_client(client=client, submissions=submissions)
    all_submissions = create_submissions_from_test_cases(submissions, test_cases)
    all_submissions = create_submissions(client=client, submissions=all_submissions)

    if wait_for_result:
        return wait(client=client, submissions=all_submissions)
    else:
        return all_submissions


def async_execute(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    source_code: Optional[str] = None,
    test_cases: Optional[Union[TestCase, TestCases]] = None,
    **kwargs,
) -> Union[Submission, Submissions]:
    return _execute(
        client=client,
        submissions=submissions,
        source_code=source_code,
        test_cases=test_cases,
        wait_for_result=False,
        **kwargs,
    )


def sync_execute(
    *,
    client: Optional[Union[Client, Flavor]] = None,
    submissions: Optional[Union[Submission, Submissions]] = None,
    source_code: Optional[str] = None,
    test_cases: Optional[Union[TestCase, TestCases]] = None,
    **kwargs,
) -> Union[Submission, Submissions]:
    return _execute(
        client=client,
        submissions=submissions,
        source_code=source_code,
        wait_for_result=True,
        test_cases=test_cases,
        **kwargs,
    )


execute = sync_execute
run = sync_execute
sync_run = sync_execute
async_run = async_execute



"""Separate file containg tests related to test case functionality."""

import judge0
import pytest
from judge0 import Status, Submission, TestCase
from judge0.api import create_submissions_from_test_cases


@pytest.mark.parametrize(
    "submissions,test_cases,expected_type",
    [
        [Submission(source_code=""), TestCase(), Submission],
        [[Submission(source_code="")], TestCase(), list],
        [Submission(source_code=""), [TestCase()], list],
        [[Submission(source_code="")], [TestCase()], list],
    ],
)
def test_create_submissions_from_test_cases_return_type(
    submissions, test_cases, expected_type
):
    output = create_submissions_from_test_cases(submissions, test_cases)
    assert type(output) == expected_type


@pytest.mark.parametrize(
    "source_code_or_submissions,test_cases,expected_status",
    [
        [
            "print(f'Hello, {input()}')",
            [TestCase("Judge0", "Hello, Judge0")],
            [Status.ACCEPTED],
        ],
        [
            "print(f'Hello, {input()}')",
            [
                TestCase("Judge0", "Hello, Judge0"),
                TestCase("pytest", "Hello, pytest"),
            ],
            [Status.ACCEPTED, Status.ACCEPTED],
        ],
        [
            Submission(source_code="print(f'Hello, {input()}')"),
            [
                TestCase("Judge0", "Hello, Judge0"),
            ],
            [Status.ACCEPTED],
        ],
        [
            Submission(source_code="print(f'Hello, {input()}')"),
            [
                TestCase("Judge0", "Hello, Judge0"),
                TestCase("pytest", "Hi, pytest"),
            ],
            [Status.ACCEPTED, Status.WRONG_ANSWER],
        ],
        [
            [
                Submission(source_code="print(f'Hello, {input()}')"),
                Submission(source_code="print(f'Hello,  {input()}')"),
            ],
            [
                TestCase("Judge0", "Hello, Judge0"),
                TestCase("pytest", "Hello, pytest"),
            ],
            [
                Status.ACCEPTED,
                Status.ACCEPTED,
                Status.WRONG_ANSWER,
                Status.WRONG_ANSWER,
            ],
        ],
    ],
)
def test_test_cases_from_run(
    source_code_or_submissions, test_cases, expected_status, request
):
    client = request.getfixturevalue("judge0_ce_client")

    if isinstance(source_code_or_submissions, str):
        submissions = judge0.run(
            client=client,
            source_code=source_code_or_submissions,
            test_cases=test_cases,
        )
    else:
        submissions = judge0.run(
            client=client,
            submissions=source_code_or_submissions,
            test_cases=test_cases,
        )

    assert [submission.status for submission in submissions] == expected_status


@pytest.mark.parametrize(
    "submissions,expected_status",
    [
        [
            Submission(
                source_code="print(f'Hello, {input()}')",
                stdin="Judge0",
                expected_output="Hello, Judge0",
            ),
            Status.ACCEPTED,
        ],
        [
            [
                Submission(
                    source_code="print(f'Hello, {input()}')",
                    stdin="Judge0",
                    expected_output="Hello, Judge0",
                ),
                Submission(
                    source_code="print(f'Hello, {input()}')",
                    stdin="pytest",
                    expected_output="Hello, pytest",
                ),
            ],
            [Status.ACCEPTED, Status.ACCEPTED],
        ],
    ],
)
def test_no_test_cases(submissions, expected_status, request):
    client = request.getfixturevalue("judge0_ce_client")

    submissions = judge0.run(
        client=client,
        submissions=submissions,
    )

    if isinstance(submissions, list):
        assert [submission.status for submission in submissions] == expected_status
    else:
        assert submissions.status == expected_status


@pytest.mark.parametrize("n_submissions", [42, 84])
def test_batched_test_cases(n_submissions, request):
    client = request.getfixturevalue("judge0_ce_client")
    submissions = [
        Submission(source_code=f"print({i})", expected_output=f"{i}")
        for i in range(n_submissions)
    ]

    results = judge0.run(client=client, submissions=submissions)

    assert len(results) == n_submissions
    assert all([result.status == Status.ACCEPTED for result in results])
